package de.hftstuttgart.unifiedticketing.systems.gitlab;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.ArrayNode;
import com.fasterxml.jackson.databind.node.ObjectNode;
import de.hftstuttgart.unifiedticketing.core.Logging;
import de.hftstuttgart.unifiedticketing.core.Ticket;
import de.hftstuttgart.unifiedticketing.core.TicketSystem;
import de.hftstuttgart.unifiedticketing.exceptions.*;
import de.hftstuttgart.unifiedticketing.exceptions.*;
import okhttp3.*;

import java.io.IOException;
import java.util.HashSet;
import java.util.Objects;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;

public class GitlabTicket extends Ticket<GitlabTicketSystem, GitlabTicket>
{
    private static Logger logger = Logging.getLogger(GitlabTicket.class.getName());

    protected GitlabTicket(GitlabTicketSystem parent)
    {
        super(parent);
    }

    /**
     * method to parse a library conform ticket instance from the JSON response of gitlab api
     * @param parent ticket system instance to be placed as parent
     * @param response api response
     * @return library conform ticket instance
     */
    protected static GitlabTicket fromTicketResponse(GitlabTicketSystem parent, GitlabTicketResponse response)
    {
        GitlabTicket ret = new GitlabTicket(parent);
        ret.description = response.description;
        ret.id = String.valueOf(response.iid);
        if (response.labels != null)
        {
            ret.labels = response.labels;
        }
        ret.open = (response.state != null)
            ? !response.state.equalsIgnoreCase("closed")
                : true;
        ret.title = response.title;

        if (response.assignees != null)
        {
            ret.assignees = response.assignees.stream()
                .map(a -> new GitlabTicketAssignee(a.id, a.name, a.username))
                .collect(Collectors.toCollection(HashSet::new));
        }

        return ret;
    }

    /**
     * adds assignee by it's user id
     *
     * @param userId gitlab user id parsable as integer
     * @return this ticket
     * @throws AssertionException if {@code userId} is not integer parsable
     */
    @Override
    public GitlabTicket addAssignee(String userId)
    {
        try
        {
            return this.addAssignee(Integer.parseInt(userId));
        } catch (NumberFormatException e)
        {
            logger.log(Level.SEVERE, String.format("%s is not applicable for a Gitlab user id", userId));
            if (this.parent.getConfigTrue(TicketSystem.ConfigurationOptions.RETURN_NULL_ON_ERROR)) return this;
            else throw new AssertionException(e);
        }
    }

    /**
     * adds assignee by it's user id
     *
     * @param userId gitlab user id
     * @return this ticket
     */
    public GitlabTicket addAssignee(int userId)
    {
        this.assignees.add(new GitlabTicketAssignee(userId, null, null));
        this.updatedFields.add(FieldNames.ASSIGNEES.name());
        return this;
    }

    /**
     * removes assignee by it's user id
     *
     * @param userId gitlab user id parsable as integer
     * @return this ticket
     * @throws AssertionException if {@code userId} is not integer parsable
     */
    @Override
    public GitlabTicket removeAssignee(String userId)
    {
        int id;
        try
        {
            return this.removeAssignee(Integer.parseInt(userId));
        } catch (NumberFormatException e)
        {
            logger.log(Level.SEVERE, String.format("%s is not applicable for a Gitlab user id", userId));
            if (this.parent.getConfigTrue(TicketSystem.ConfigurationOptions.RETURN_NULL_ON_ERROR)) return this;
            else throw new AssertionException(e);
        }
    }

    /**
     * removes assignee by it's user id
     *
     * @param userId gitlab user id parsable as integer
     * @return this ticket
     */
    public GitlabTicket removeAssignee(int userId)
    {
        this.assignees.removeIf(a -> Objects.equals(a.id, String.valueOf(userId)));
        this.updatedFields.add(FieldNames.ASSIGNEES.name());
        return this;
    }

    protected OkHttpClient getHttpClient()
    {
        return new OkHttpClient();
    }

    /**
     * checks for changed fields and serializes them into a PUT request
     * to update this ticket instance.
     * @return newly received ticket data as new {@link GitlabTicket instance}
     */
    @Override
    public GitlabTicket save()
    {
        if (updatedFields.size() == 0)
        {
            logger.info("No changed fields, no save required");
            return this;
        }

        OkHttpClient client = this.getHttpClient();
        ObjectMapper mapper = new ObjectMapper()
            .configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);

        logger.log(Level.FINEST, String.format(
            "[Ticket %s] preparing body for update request",
            this.id
        ));
        ObjectNode body = mapper.createObjectNode();
        for (String name : this.updatedFields)
        {
            if (FieldNames.ASSIGNEES.name().equals(name))
            {
                ArrayNode ids = body.putArray("assignee_ids");
                try
                {
                    this.assignees.forEach(a -> ids.add(Integer.parseInt(a.id)));
                } catch (NumberFormatException e)
                {
                    logger.log(Level.SEVERE, "parsing assignee id failed on serialization");
                    if (this.parent.getConfigTrue(TicketSystem.ConfigurationOptions.RETURN_NULL_ON_ERROR))
                    {
                        body.remove("assignee_ids");
                        continue;
                    }
                    else
                    {
                        throw new SerializationException(e);
                    }
                }
            }
            else if (FieldNames.DESCRIPTION.name().equals(name))
            {
                body.put("description", this.description);
            }
            else if (FieldNames.LABELS.name().equals(name))
            {
                body.put("labels", this.labels.stream().reduce((l1, l2) -> l1 + "," + l2).orElse(""));
            }
            else if (FieldNames.OPEN.name().equals(name))
            {
                body.put("state_event", (this.open) ? "reopen" : "close");
            }
            else if (FieldNames.TITLE.name().equals(name))
            {
                body.put("title", this.title);
            }
            else
            {
                logger.log(Level.WARNING, String.format(
                    "[Ticket %s] unknown field %s will be ignored from update",
                    this.id,
                    name
                ));
                continue;
            }

            logger.log(Level.FINEST, String.format(
                "[Ticket %s] %s added to update",
                this.id,
                name
            ));
        }

        logger.log(Level.FINEST, String.format(
            "[Ticket %s] request body for update prepared",
            this.id
        ));

        Request.Builder builder = new Request.Builder()
                .url(String.format("%s/issues/%s", this.parent.baseUrl, this.id));
        try
        {
            logger.log(Level.FINEST, String.format(
                "[Ticket %s] serializing update request body",
                this.id
            ));

            String bodyJson = mapper.writeValueAsString(body);

            logger.log(Level.FINEST, String.format(
                "[Ticket %s] serialized JSON:\n%s",
                this.id,
                bodyJson
            ));
            builder
                .put(RequestBody.create(bodyJson, MediaType.get("application/json")));
        } catch (JsonProcessingException e)
        {
            logger.log(Level.SEVERE, String.format(
                "[Ticket %s] serializing update request body FAILED!",
                this.id
            ));

            if (this.parent.getConfigTrue(TicketSystem.ConfigurationOptions.RETURN_NULL_ON_ERROR)) return null;
            else throw new SerializationException(e);
        }

        if (parent.apiKey != null)
        {
            builder.addHeader("PRIVATE-TOKEN", parent.apiKey);
            logger.log(Level.FINEST, String.format(
                "[Ticket %s] added token authentication header",
                this.id
            ));
        }

        Response response;
        try
        {
            Request request = builder.build();
            logger.log(Level.FINEST, String.format(
                "[Ticket %s] created request:\n%s",
                this.id,
                (this.parent.apiKey != null)
                    ? request.toString().replace(this.parent.apiKey, "SECRET")
                    : request.toString()
            ));

            response = client.newCall(request).execute();
        }
        catch (IOException e)
        {
            logger.log(Level.SEVERE, String.format(
                "[Ticket %s] update request FAILED",
                this.id
            ));

            if (this.parent.getConfigTrue(TicketSystem.ConfigurationOptions.RETURN_NULL_ON_ERROR)) return null;
            else throw new HttpReqeustException(e);
        }

        if (response.code() >= 400)
        {
            logger.log(Level.SEVERE, String.format(
                "[Ticket %s] update request failed with response code %d",
                this.id,
                response.code()
            ));

            if (this.parent.getConfigTrue(TicketSystem.ConfigurationOptions.RETURN_NULL_ON_ERROR)) return null;
            else throw new HttpResponseException(
                String.format("ticket save failed, error response code: %d", response.code()),
                response.code());
        }

        logger.log(Level.FINEST, String.format(
            "[Ticket %s] response received",
            this.id
        ));

        ResponseBody responseBody;
        responseBody = response.body();

        if (responseBody == null)
        {
            logger.log(Level.SEVERE, String.format(
                "[Ticket %s] update request didn't deliver a body in response",
                this.id
            ));

            if (this.parent.getConfigTrue(TicketSystem.ConfigurationOptions.RETURN_NULL_ON_ERROR)) return null;
            else throw new HttpResponseException("ticket save failed, no response body", response.code());
        }

        GitlabTicketResponse ticketResponse;
        try
        {
            ticketResponse = mapper.readValue(responseBody.bytes(), GitlabTicketResponse.class);

            logger.log(Level.FINEST, String.format(
                "[Ticket %s] parsed response body to ticketResponse instance",
                this.id
            ));
        } catch (IOException e)
        {
            logger.log(Level.SEVERE, String.format("parsing update response FAILED with: %s", e.getMessage()));

            if (this.parent.getConfigTrue(TicketSystem.ConfigurationOptions.RETURN_NULL_ON_ERROR)) return null;
            else throw new DeserializationException(e);
        }

        this.updatedFields.clear();

        logger.log(Level.FINEST, String.format(
            "[Ticket %s] update mark state reset",
            this.id
        ));

        return GitlabTicket.fromTicketResponse(this.parent, ticketResponse);
    }

    /**
     * compares a given Object to this one, including all data fields.
     * The normal {@link #equals(Object)} method does only compare Ticket Type and id,
     * but not the data fields like this one.
     * @param o Object to compare to this
     * @return true if the given object is equal in terms of type, id and content to this
     */
    public boolean deepEquals(Object o)
    {
        if (!equals(o)) return false;
        GitlabTicket t = (GitlabTicket) o;

        return Objects.equals(title, t.title)
            && Objects.equals(description, t.description)
            && Objects.equals(labels, t.labels)
            && Objects.equals(assignees, t.assignees);
    }
}
