package de.hftstuttgart.unifiedticketing.systems.gitlab;

import de.hftstuttgart.unifiedticketing.core.TicketSystem;
import de.hftstuttgart.unifiedticketing.exceptions.AssertionException;
import de.hftstuttgart.unifiedticketing.exceptions.UnifiedticketingException;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mockito;

import java.util.Collections;
import java.util.LinkedList;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

public class GitlabTicketSystemTest
{
    public GitlabTicketSystem instance;

    @BeforeEach
    public void initBeforeEach()
    {
        instance = new GitlabTicketSystem();
    }

    @Test
    public void testFromUri()
    {
        String base = "gitlab.com";
        int projectId = 23554234;
        String apikey = "afdaf3aqraf3afafmyxcbvmyxvbas3wrawra";
        TicketSystem actual = TicketSystem.fromUri(buildTestUri(true, base, projectId, apikey));

        assertEquals(buildFinalBaseurl(true, base, projectId), actual.baseUrl);
        assertEquals(apikey, actual.apiKey);
        assertEquals(GitlabTicketSystem.class, actual.getClass());

        base = "gitlab.somedomain.tld:8080";
        projectId = 5;
        apikey = "asfewfa3r2rfa34fraf";
        actual = TicketSystem.fromUri(buildTestUri(false, base, projectId, apikey));

        assertEquals(buildFinalBaseurl(false, base, projectId), actual.baseUrl);
        assertEquals(apikey, actual.apiKey);
        assertEquals(GitlabTicketSystem.class, actual.getClass());

        actual = TicketSystem.fromUri(buildTestUri(false, base, projectId, null));
        assertEquals(buildFinalBaseurl(false, base, projectId), actual.baseUrl);
        assertNull(actual.apiKey);
        assertEquals(GitlabTicketSystem.class, actual.getClass());

        assertThrows(AssertionException.class, () -> TicketSystem.fromUri("unifiedticketing:gitlab:blablabal"));
    }

    @Test
    public void testCreateTicket()
    {
        assertEquals(GitlabTicketBuilder.class, instance.createTicket().getClass());
        assertSame(instance, instance.createTicket().parent);
    }

    @Test
    public void testFind()
    {
        assertEquals(GitlabFilter.class, instance.find().getClass());
        assertSame(instance, instance.find().parent);
    }

    @Test
    public void testGetTicketById()
    {
        String ticketId = "5";
        GitlabFilter filter = Mockito.spy(instance.find());
        instance = spy(instance);
        doReturn(filter).when(instance).find();
        doReturn(null).when(filter).get();

        assertNull(instance.getTicketById(ticketId));
        verify(instance, times(1)).find();
        verify(filter, times(1)).withId(ticketId);
        verify(filter, times(1)).get();

        doReturn(new LinkedList<>()).when(filter).get();
        assertThrows(UnifiedticketingException.class, () -> instance.getTicketById(ticketId));

        GitlabTicket ticket = new GitlabTicket(instance);
        doReturn(new LinkedList<>(Collections.singleton(ticket))).when(filter).get();
        assertSame(ticket, instance.getTicketById(ticketId));
    }

    @Test
    public void testSupport()
    {
        assertAll(
            () -> assertTrue(instance.hasAssigneeSupport()),
            () -> assertTrue(instance.hasDefaultPagination()),
            () -> assertTrue(instance.hasLabelSupport()),
            () -> assertTrue(instance.hasPaginationSupport()),
            () -> assertTrue(instance.hasReturnNullOnErrorSupport())
        );
    }

    private static String buildTestUri(boolean https, String base, int projectId, String apikey)
    {
        return String.format(
            "unifiedticketing:gitlab:%s://%s:%d%s",
            (https) ? "https": "http",
            base,
            projectId,
            (apikey == null) ? "" : ":" + apikey);
    }

    private static String buildFinalBaseurl(boolean https, String base, int projectId)
    {
        return String.format("%s://%s/api/v4/projects/%d", (https) ? "https" : "http", base, projectId);
    }
}
